/*
#             (C) 2008 Lukas karas <lukas.karas@centrum.cz>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation; either version 2.1 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

/*
 Classic model allow control video device only by application, 
 that read stream from its.
 
 But a lot of applications (for example Skype for Linux) not support all posible controls.
 
 This lib allow second applicatios send CTLs to video device via shared memory.
*/


#include <linux/videodev2.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <sys/wait.h>

#include "../include/v4lcontrol.h"

/**
 Get maximal and latest CIDs using by device driver.
 Driver should support V4L2_CTRL_FLAG_NEXT_CTRL for get correct values!
*/
void get_last_and_max_ctrl_id(int fd, long unsigned int *max,long unsigned int *last){
	int i;
	struct v4l2_queryctrl   qctrl = {
		.id = V4L2_CID_BASE
	};
	*max = V4L2_CID_BASE;
	*last = V4L2_CID_BASE;	
	
	/* find first CID */
	i = V4L2_CID_BASE;
	for (qctrl.id = V4L2_CID_BASE; qctrl.id < V4L2_CID_LASTP1;qctrl.id++){
		if (v4l2_ioctl(fd,VIDIOC_QUERYCTRL,&qctrl) >= 0 ){
			i = qctrl.id;
			break;
		}
	}
	qctrl.id = i;
	
	for (i = 0; i < V4L2_CTRL_ID_MASK; i++) {
		if (i>0){
			qctrl.id |= V4L2_CTRL_FLAG_NEXT_CTRL;
		}
			
		if (0 > v4l2_ioctl(fd,VIDIOC_QUERYCTRL,&qctrl)){
			break;
		}
		if (qctrl.flags & V4L2_CTRL_FLAG_DISABLED)
			continue;
		
		*last = qctrl.id;
		if (*last > *max)
				*max = * last;
	}
}

/** 
 Decrease value of ipc semaphore
*/
int binary_semaphore_down (int semid, int counterId){
	struct sembuf operation;
	operation.sem_num = counterId;
	operation.sem_op = -1;
	operation.sem_flg = 0;
	return semop (semid, &operation, 1);
}

/** 
 Increase value of ipc semaphore
 */
int binary_semaphore_up(int semid, int counterId){
	struct sembuf operation;
	operation.sem_num = counterId;
	operation.sem_op = 1;
	operation.sem_flg = 0;
	return semop (semid, &operation, 1);
}

void *event_thread_func(void *arg){
	int fd = ((struct thread_arg *)arg)->fd;
	struct v4lcontrol_data *control_data;
	key_t data_key = ftok( (char *)((struct thread_arg *)arg)->device, SHM_DATA_ID);
	key_t semaphores_key = ftok( (char *)((struct thread_arg *)arg)->device, SHM_SEMAPTHORES_ID);
	unsigned char exit = 0;
	struct v4lcontrol_shared *data;
	int i;
	
	control_data = ((struct thread_arg *)arg)->control_data;
		
	if (( control_data->shm_semaphores_id = semget(semaphores_key, SEMAPHORES_COUNT, IPC_CREAT | 0600) ) < 0) {
		V4L2_LOG_ERR("shmget semaphores");
	}
	
	semctl(control_data->shm_semaphores_id, SEMAPHORE_CLIENT_SEM_ID, SETVAL, 1);
	semctl(control_data->shm_semaphores_id, SEMAPHORE_LIB_DATA_SEM_ID, SETVAL, 0);
	semctl(control_data->shm_semaphores_id, SEMAPHORE_CLIENT_DATA_SEM_ID, SETVAL, 0);
		
	if (( control_data->shm_data_id = shmget(data_key, sizeof( struct v4lcontrol_shared ), IPC_CREAT | 0600)  ) < 0) {
		V4L2_LOG_ERR("shmget data");
	}
	
	
	if ((data = (struct v4lcontrol_shared *)shmat(control_data->shm_data_id, NULL, 0)) == (struct v4lcontrol_shared *) -1) {
		V4L2_LOG_ERR("shmat");
	}
	
	data->lib_pid = (int) getpid ();
	data->ioctl_request = CTL_EXIT;
	
	/* do ctls */
	while(!exit){
		binary_semaphore_down(control_data->shm_semaphores_id, SEMAPHORE_LIB_DATA_SEM_ID);
		if (v4l2_log_file)
			fprintf(v4l2_log_file,"data recieved (shm)... (ioctl 0x%x)\n", (int)data->ioctl_request);
		switch( data->ioctl_request ){
			case CTL_EXIT:
				data->result = 0;
				exit = 1;
				break;
			default:
				data->result = v4l2_ioctl(fd,data->ioctl_request,&data->arg);
		}
		binary_semaphore_up(control_data->shm_semaphores_id, SEMAPHORE_CLIENT_DATA_SEM_ID);
	}
	binary_semaphore_up(control_data->shm_semaphores_id, SEMAPHORE_CLIENT_DATA_SEM_ID);
	
	
	if ( shmdt( data ) < 0 ) {
		V4L2_LOG_ERR("shmdt");
	}
	
	if ( shmctl(control_data->shm_data_id, IPC_RMID, 0 ) < 0 ) {
		V4L2_LOG_ERR("shmctl data");
	}
	
	if ( shmctl(control_data->shm_semaphores_id, IPC_RMID, 0 ) < 0 ) {
		V4L2_LOG_ERR("shmctl semaphores");
	}
	
	return (void *)NULL;
}


struct v4lcontrol_data *v4lcontrol_init(int fd, const char *device){
	int rc;
	pthread_attr_t attr;
	struct v4lcontrol_data *data = calloc(1, sizeof(struct v4lcontrol_data));
	struct thread_arg arg = {
		.fd = fd,
		.device = device,
		.control_data = data
	};
	
	long unsigned int max;
	
	memcpy(data, &v4lcontrol_data_skeleton, sizeof(struct v4lcontrol_data));
	
	/* assign free private CIDs for software controls */
	get_last_and_max_ctrl_id(fd, &max, &data->last_ctrl_id);
	if (max < V4L2_CID_PRIVATE_BASE)
		max = V4L2_CID_PRIVATE_BASE -1;
	
	data->sw_ctrl[SW_CTRL_WB_INDEX].id 		= max + 1;
	data->sw_ctrl[SW_CTRL_ROTATE_INDEX].id 	= max + 2;
	
	/* init event thread */
	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);    
	
	rc = pthread_create(&data->event_thread, &attr, event_thread_func, (void *)&arg);
	if (rc){ 
		V4L2_LOG_ERR("Error in function pthread_create()"); 
	}
	
	/* Free attribute */
	pthread_attr_destroy(&attr);
	
	return data;
}


int v4lcontrol_destroy(struct v4lcontrol_data *control){
	int rc;
	struct v4lcontrol_shared *data;	
	
	binary_semaphore_down(control->shm_semaphores_id, SEMAPHORE_CLIENT_SEM_ID);
	
	if ((data = (struct v4lcontrol_shared *)shmat(control->shm_data_id, NULL, 0)) == (struct v4lcontrol_shared *) -1) {
		V4L2_LOG_ERR("shmat");
	}
	data->ioctl_request = CTL_EXIT;
	if ( shmdt( data ) < 0 ) {
		V4L2_LOG_ERR("shmdt");
		return -1;
	}
	
	binary_semaphore_up(control->shm_semaphores_id, SEMAPHORE_LIB_DATA_SEM_ID);
	binary_semaphore_down(control->shm_semaphores_id, SEMAPHORE_CLIENT_DATA_SEM_ID);
	
	binary_semaphore_up(control->shm_semaphores_id, SEMAPHORE_CLIENT_SEM_ID);
	
	/* wait for end of thread */
	rc = pthread_join(control->event_thread, NULL );
	if (rc){
		V4L2_LOG_ERR("ERROR return code from pthread_join() is %d\n", rc);
	}

	free(control);
	return 0;
}

int v4lcontrol_queryctrl(int fd, struct v4l2_queryctrl *qctrl){
	int result = -1; 
	int i, index;
	unsigned char next = 0;
	
	/* first, we try software ctrls */
	index = v4l2_get_index(fd);
	
	if ((qctrl->id ^ V4L2_CTRL_FLAG_NEXT_CTRL) == devices[index].control->last_ctrl_id ){
		next = 1;
	}
	if (index!= -1){
		for (i=0; i < COUNT_SW_CTRL; i++){
			if (devices[index].control->sw_ctrl[i].id == qctrl->id || next ){
				/* software ctrl is selected */
				memcpy(qctrl, &devices[index].control->sw_ctrl[i], sizeof(struct v4l2_queryctrl));
				return 0;
			}
			if ((qctrl->id ^ V4L2_CTRL_FLAG_NEXT_CTRL) == devices[index].control->sw_ctrl[i].id){
				next = 1;
			}
		}
	}
	
	/* driver (hardware) ctrl is selected */
	result = syscall(SYS_ioctl, fd, VIDIOC_QUERYCTRL, qctrl);
	v4l2_log_ioctl(VIDIOC_QUERYCTRL, qctrl, result);
	return result;	
}

int v4lcontrol_setctrl(int fd, struct v4l2_control *ctrl){
	
	struct v4l2_queryctrl qctrl = { .id = ctrl->id };
	int result, i, index;
	
	/* first, try software ctrls */
	index = v4l2_get_index(fd);
	if (index!= -1){
		for (i=0; i < COUNT_SW_CTRL; i++){
			if (devices[index].control->sw_ctrl[i].id == ctrl->id){
				devices[index].control->value[i] = ctrl->value;
				return 0;
			}
		}
	}
	
	if (result = v4lcontrol_queryctrl(fd, &qctrl))
		return result;

	if (!(qctrl.flags & V4L2_CTRL_FLAG_DISABLED) &&
			 !(qctrl.flags & V4L2_CTRL_FLAG_GRABBED)) {
		if (qctrl.type == V4L2_CTRL_TYPE_BOOLEAN)
			ctrl->value = ctrl->value? 1:0;

		result = syscall(SYS_ioctl, fd, VIDIOC_S_CTRL, ctrl);
	}

	return result;
}

int v4lcontrol_getctrl(int fd, struct v4l2_control *ctrl){
	
	struct v4l2_queryctrl qctrl = { .id = ctrl->id };
	int i, index, result;

	/* first, we try software ctrls */
	index = v4l2_get_index(fd);
	if (index!= -1){
		for (i=0; i < COUNT_SW_CTRL; i++){
			if (devices[index].control->sw_ctrl[i].id == ctrl->id){
				ctrl->value = devices[index].control->value[i];
				return 0;
			}
		}
	}
		
	if (v4lcontrol_queryctrl(fd, &qctrl))
		return 0;
	
	if (qctrl.flags & V4L2_CTRL_FLAG_DISABLED)
		return 0;

	return  syscall(SYS_ioctl, fd, VIDIOC_G_CTRL, ctrl);
}
